

class Book:
    def __init__(self, data, **kwargs):
        self.name = 'blueprint-book'
        self.label = ''
        self.label_color = None
        self.blueprints = []
        self.active_index = 0
        self.version = 0



from py_factorio_blueprints.util import \
    Vector, NameStr, Connection, Direction, ControlBehaviorMeta
from py_factorio_blueprints.entity_mixins import \
    BaseMixin, SignalName, Base
from py_factorio_blueprints.exceptions import *


class PositionField:
    def __set_name__(self, owner, name):
        self.name = "__" + name

    def __set__(self, instance, value):
        setattr(instance, self.name, Vector(value))

    def __get__(self, instance, owner):
        return getattr(instance, self.name, Vector(0, 0))


class DirectionField:
    def __set_name__(self, owner, name):
        self.name = "__" + name

    def __set__(self, instance, value):
        setattr(instance, self.name, Direction(value))

    def __get__(self, instance, owner):
        return getattr(instance, self.name, Direction(0))


class ItemName(Base):
    class NameStr(str):
        @property
        def data(self):
            from py_factorio_blueprints import Blueprint
            return Blueprint.entity_prototypes[self]

    def __set_name__(self, owner, name):
        self.name = "__" + name

    def __set__(self, instance, value):
        if not getattr(instance, 'strict', True):
            setattr(instance, self.name, value)
            return

        from py_factorio_blueprints import Blueprint

        if value not in Blueprint.item_prototypes:
            raise UnknownItem
        setattr(instance, self.name, value)

    def __get__(self, instance, owner):
        return ItemName.NameStr(getattr(instance, self.name, ""))


class EntityName(Base):
    class NameStr(str):
        @property
        def prototype(self):
            return self.data['type']

        @property
        def data(self):
            from py_factorio_blueprints.blueprint import Blueprint
            return Blueprint.entity_prototypes[self]

        @property
        def selection_box(self):
            return (
                Vector(**self.data['selection_box']['left_top']),
                Vector(**self.data['selection_box']['right_bottom'])
            )

    def __set_name__(self, owner, name):
        self.name = name
        self._name = "__" + name

    def __set__(self, instance, value):
        if not getattr(instance, 'strict', True):
            setattr(instance, self._name, value)
            return

        from py_factorio_blueprints.blueprint import Blueprint

        if value not in Blueprint.entity_prototypes:
            raise UnknownEntity(value)
        setattr(instance, self._name, value)
        prototype = Blueprint.entity_prototypes[value]['type']

        if instance.blueprint:
            custom_prototypes = instance.blueprint.custom_entity_prototypes\
                .get(prototype, {})
            custom_mixins = custom_prototypes.get('mixins', [])
        else:
            custom_mixins = []
        from py_factorio_blueprints.entity_prototypes import entity_prototypes
        mixins = entity_prototypes[prototype].get('mixins', [])
        instance.add_mixins(*custom_mixins, *mixins)

    def __get__(self, instance, owner):
        return EntityName.NameStr(getattr(instance, self._name, ""))


class Entity(BaseMixin):
    ID = 'entity_number'
    name = EntityName()
    position = PositionField()
    direction = DirectionField()

    def __repr__(self):
        return f'<Entity (name: "{self.name}", position: ' \
               f'{self.position}, direction: {self.direction})>'

    def __str__(self):
        return f'<Entity (name: "{self.name}")>'

    def set_mixins(self, *mixins):
        """ Resets self to the base Entity class and adds the given mixins """
        self.__class__ = type('Entity', (Entity, *mixins), {})

    def add_mixins(self, *mixins):
        """ Adds additional mixins to the current self class """
        self.__class__ = type('Entity', (self.__class__, *mixins), {})

    def __init__(self, *args, name, position, strict=True,
                 blueprint_layer=None,
                 direction=0, entity_number=None, **kwargs):
        self.strict = strict
        self._blueprint_layer = blueprint_layer
        self.entity_number = entity_number
        self._auto_entity_number = entity_number
        self.name = name
        self.position = position
        self.direction = direction

        self.raw_connections = kwargs.pop('connections', None)
        # self.connections = []

        if 'control_behavior' in kwargs:
            self.control_behavior = self.ControlBehavior(
                _entity=self,
                **kwargs.pop('control_behavior'))

        super().__init__(*args, **kwargs)

    @property
    def blueprint(self):
        try:
            return self._blueprint_layer.blueprint
        except AttributeError:
            return None

    def get_connections(self):
        connections = [connection.orientate(self)
                       for connection in self.blueprint.connections
                       if connection.attached_to(self)]
        return sorted(
            connections, key=lambda x: (
                x.from_entity._auto_entity_number,
                x.to_entity._auto_entity_number))

    def connect(
            self,
            to_entity,
            from_side='in',
            to_side='in',
            color='red'):
        conn = Connection(self, to_entity, from_side, to_side, color)
        if conn not in self.blueprint.connections:
            self.blueprint.connections.append(conn)
        return conn

    def connections_to_json(self):
        connections = self.get_connections()
        obj = {}
        for connection in connections:
            if connection.from_side not in obj:
                obj[connection.from_side] = {}
            if connection.color not in obj[connection.from_side]:
                obj[connection.from_side][connection.color] = []
            if connection.to_entity is self:
                if connection.to_side not in obj:
                    obj[connection.to_side] = {}
                if connection.color not in obj[connection.to_side]:
                    obj[connection.to_side][connection.color] = []
                result = {
                    'entity_id': self._auto_entity_number
                }
                if self.NR_CONNECTIONS == 2:
                    result['circuit_id'] = connection.from_side
                obj[connection.to_side][connection.color].append(result)

            result = {
                'entity_id': connection.to_entity._auto_entity_number,
            }
            if connection.to_entity.NR_CONNECTIONS == 2:
                result['circuit_id'] = connection.from_side
            obj[connection.from_side][connection.color].append(result)
        return obj

    @property
    def top_left(self):
        top_left, bottom_right = self.selection_box
        return self.position + top_left

    @property
    def top_right(self):
        top_left, bottom_right = self.selection_box
        return self.position + Vector(bottom_right.x, top_left.y)

    @property
    def bottom_left(self):
        top_left, bottom_right = self.selection_box
        return self.position + Vector(top_left.x, bottom_right.y)

    @property
    def bottom_right(self):
        top_left, bottom_right = self.selection_box
        return self.position + bottom_right

    @property
    def selection_box(self):
        selection_box = self.name.selection_box
        def r(box):
            top_left, bottom_right = box
            return top_left.yx, bottom_right.yx

        for _ in range(self.direction // 2):
            selection_box = r(selection_box)

        return selection_box

    def collides(self, position):
        top_left, bottom_right = self.selection_box
        position -= self.position
        return top_left < position < bottom_right

    def rotate(self, amount,
               around=Vector(0, 0), direction=Direction.CLOCKWISE):
        position = self.position - around
        amount %= 4
        if direction != Direction.CLOCKWISE:
            amount %= 4 - amount

        def r(v):
            return Vector(-v.y, v.x)

        for _ in range(amount):
            position = r(position)
        self.position = position + around

        mixin = super()
        if hasattr(mixin, "rotate"):
            mixin.rotate(amount)

    def to_json(self, obj=None):
        if obj is None:
            obj = {}
        obj['entity_number'] = self._auto_entity_number
        obj['name'] = self.name
        obj['position'] = self.position.to_json()
        connections = self.connections_to_json()
        if connections != {}:
            obj['connections'] = connections
        return super().to_json(obj)



import logging
from py_factorio_blueprints import util
from py_factorio_blueprints.entity import Entity as BaseEntity
from py_factorio_blueprints.exceptions import *
from py_factorio_blueprints.util import (
    Color, Tile, Connection, Vector, obj_set, Direction
)
from py_factorio_blueprints.entity_mixins import SignalName
import json


logger = logging.getLogger('py_factorio_blueprints.blueprint')


class Schedule:
    def __init__(self, blueprint, *, schedule, locomotives):
        self.blueprint = blueprint
        self.locomotives = self.find_locomotives(locomotives)
        self.schedule = schedule

    def find_locomotives(self, locomotives):
        locos = [
            self.blueprint.entities.get_by_id(loco_id)
            for loco_id in locomotives]
        return [loco for loco in locos if loco is not None]

    def to_json(self):
        return {
            'locomotives': [
                loco._auto_entity_number
                for loco in self.locomotives],
            'schedule': self.schedule
        }


class BlueprintLayer:
    def __set_name__(self, owner, name):
        self.name = name
        self.owner = owner

    def __init__(self, blueprint, obj_type, strict=True):
        self.__blueprint = blueprint
        self.strict = strict
        self.obj_type = obj_type
        self.objs = []

    def __iter__(self):
        yield from self.objs

    def __getitem__(self, vector):
        if type(vector) is tuple:
            vector = Vector(vector)

        return [obj for obj in self.objs
                if obj.collides(vector)]

    def __delitem__(self, obj):
        if isinstance(obj, self.obj_type):
            self.objs.remove(obj)
            obj._blueprint_layer = None
        else:
            objs = self[obj]
            for obj in objs:
                self.objs.remove(obj)
                obj._blueprint_layer = None

    def add(self, obj):
        if not isinstance(obj, self.obj_type):
            raise TypeError(
                f"{obj} is not of type {self.obj_type}")
        if obj._blueprint_layer is not None and obj._blueprint_layer != self:
            raise DuplicateEntity(
                "Can't add Entity instance to more than one BlueprintLayer")
        if obj in self.objs:
            return
        self.objs.append(obj)
        obj._blueprint_layer = self
        self.sort()

    def remove(self, obj):
        if not isinstance(obj, self.obj_type):
            raise TypeError(
                f"{obj} is not of type {self.obj_type}")
        self.objs.remove(obj)
        obj._blueprint_layer = None

    def make(self, *args, **kwargs):
        obj = self.obj_type(*args, blueprint_layer=self, **kwargs)
        self.add(obj)
        return obj

    def sort(self):
        self.__sort()
        self.__reindex()

    def _load(self, *args, **kwargs):
        obj = self.obj_type(*args, blueprint_layer=self, **kwargs)
        self.objs.append(obj)

    @property
    def blueprint(self):
        return self.__blueprint

    def rotate(self, amount, **kwargs):
        for obj in self.objs:
            obj.rotate(amount, **kwargs)
        self.sort()

    def to_json(self):
        obj = [
            obj.to_json()
            for obj in self
        ]
        if not obj:
            return None
        return obj

    def get_by_id(self, obj_id):
        if not hasattr(self.obj_type, 'ID'):
            return None
        for obj in self.objs:
            if getattr(obj, self.obj_type.ID) == obj_id:
                return obj
        return None

    def __get__(self, instance, owner):
        return self

    def __sort(self):
        self.objs.sort(key=lambda obj: (obj.position.y, obj.position.x))

    def __reindex(self):
        for i, obj in enumerate(self.objs):
            obj._auto_entity_number = i + 1


class Blueprint:
    entity_prototypes = {}
    recipe_prototypes = {}
    item_prototypes = {}
    signal_prototypes = {}
    tile_prototypes = {}

    @classmethod
    def set_entity_prototype_data(cls, data, append=False):
        if append:
            data = {**cls.entity_prototypes, **data}
        cls.entity_prototypes = data

    @classmethod
    def set_recipe_prototype_data(cls, data, append=False):
        if append:
            data = {**cls.recipe_prototypes, **data}
        cls.recipe_prototypes = data

    @classmethod
    def set_item_prototype_data(cls, data, append=False):
        if append:
            data = {**cls.item_prototypes, **data}
        cls.item_prototypes = data

    @classmethod
    def set_signal_prototype_data(cls, data, append=False):
        if append:
            data = {**cls.signal_prototypes, **data}
        cls.signal_prototypes = data

    @classmethod
    def set_tile_prototype_data(cls, data, append=False):
        if append:
            data = {**cls.tile_prototypes, **data}
        cls.tile_prototypes = data

    @classmethod
    def import_prototype_data(cls, filename, **kwargs):
        with open(filename) as f:
            data = json.load(f)
            cls.set_entity_prototype_data(data['entity'], **kwargs)
            cls.set_recipe_prototype_data(data['recipe'], **kwargs)
            cls.set_item_prototype_data(data['item'], **kwargs)
            cls.set_signal_prototype_data(data['signal'], **kwargs)
            cls.set_tile_prototype_data(data['tile'], **kwargs)

    def __init__(self, string=None, data=None,
                 *, custom_entity_prototypes=None, strict=True,
                 verbose=False, **kwargs):
        super().__init__(**kwargs)
        self._verbose = verbose
        self.strict = strict
        self.__entities = BlueprintLayer(self, BaseEntity, strict=strict)
        self.__tiles = BlueprintLayer(self, Tile, strict=strict)
        self.item = 'blueprint'
        self.label = ''
        self.label_color = None
        self.icons = []
        self.version = 0
        self.connections = []
        self.schedules = []

        if custom_entity_prototypes is None:
            custom_entity_prototypes = {}
        self.custom_entity_prototypes = custom_entity_prototypes

        logger.debug(string)
        if string is not None:
            data = util.decode(string)
            logger.debug(data)
        if data is not None:
            self.load(data)

    def __eq__(self, other):
        if not isinstance(other, Blueprint):
            return NotImplemented
        return self.to_string() == other.to_string()

    @property
    def entities(self):
        return self.__entities

    @property
    def tiles(self):
        return self.__tiles

    def load(self, data):
        if 'blueprint' in data:
            data = data['blueprint']
        logger.debug(data)

        self.item = data.get('item', 'blueprint')
        self.label = data.get('label', None)
        label_color = data.get('label_color', None)
        if label_color is not None:
            self.label_color = Color(**label_color)
        self.version = data.get('version', 0)

        self.icons = [None, None, None, None]

        class Icon:
            name = SignalName()

        for icon in data.get('icons', []):
            logger.debug('icon:', icon)
            new_icon = Icon()
            new_icon.name = icon['signal']
            self.icons[icon['index'] - 1] = new_icon
        logger.debug(self.icons)

        for entity in data.get('entities', []):
            self.entities._load(**entity)
        self.entities.sort()

        self.parse_connections()

        for tile in data.get('tiles', []):
            self.tiles._load(**tile)
        self.tiles.sort()

        self.schedules = [
            Schedule(self, **schedule)
            for schedule in data.pop('schedules', [])
        ]

    def set_label(self, label, color=None):
        self.label = label
        if isinstance(color, Color):
            self.label_color = color
        elif color is not None:
            self.label_color = Color(**color)

    def rotate(self, amount, **kwargs):
        self.entities.rotate(amount, **kwargs)
        self.tiles.rotate(amount, **kwargs)

    def to_json(self):
        obj = {
            'item': self.item,
        }
        obj_set(obj, 'label', self.label)
        obj_set(obj, 'label_color', self.label_color)
        obj_set(obj, 'entities', self.entities.to_json())
        obj_set(obj, 'tiles', self.tiles.to_json())
        obj['icons'] = [
            {'index': i + 1, 'signal': icon.name.to_json()}
            for i, icon in enumerate(self.icons)
            if icon is not None]
        obj['version'] = self.version
        schedules = [
            schedule.to_json()
            for schedule in self.schedules]
        if schedules:
            obj['schedules'] = schedules

        return {'blueprint': obj}

    def to_json_string(self):
        return json.dumps(self.to_json())

    def to_string(self):
        obj = self.to_json()
        return util.encode(obj)

    def get_entity_by_id(self, entity_number):
        for entity in self.entities:
            if entity.entity_number == entity_number:
                return entity

    @property
    def maximum_values(self):
        maxx, minx, maxy, miny =\
            float('-inf'), float('inf'), float('-inf'), float('inf')
        for entity in self.entities:
            top_left, bottom_right = entity.top_left, entity.bottom_right
            _maxx, _minx, _maxy, _miny =\
                bottom_right.x, top_left.x, bottom_right.y, top_left.y
            if _maxx > maxx:
                maxx = int(_maxx)
            if _minx < minx:
                minx = int(_minx)
            if _maxy > maxy:
                maxy = int(_maxy)
            if _miny < miny:
                miny = int(_miny)
        return maxx, minx, maxy, miny

    @property
    def center(self):
        maxx, minx, maxy, miny = self.maximum_values
        return Vector((maxx + minx) / 2, (maxy + miny) / 2)

    @property
    def weighted_center(self):
        center = Vector(0, 0)
        for entity in self.entities:
            center += entity.position
        center /= len(self.entities)
        return center

    @property
    def top_left(self):
        maxx, minx, maxy, miny = self.maximum_values
        return Vector(minx, miny)

    @property
    def top_right(self):
        maxx, minx, maxy, miny = self.maximum_values
        return Vector(maxx, miny)

    @property
    def bottom_left(self):
        maxx, minx, maxy, miny = self.maximum_values
        return Vector(minx, maxy)

    @property
    def bottom_right(self):
        maxx, minx, maxy, miny = self.maximum_values
        return Vector(maxx, maxy)

    @property
    def corners(self):
        maxx, minx, maxy, miny = self.maximum_values
        return (
            Vector(minx, miny), Vector(maxx, miny),
            Vector(minx, maxy), Vector(maxx, maxy))

    def recenter(self, around=None):
        if around is not None:
            center = around
        else:
            center = self.center.ceil()
        if center == Vector(0, 0):
            return
        for entity in self.entities:
            entity.position -= center
        self.replace_entities()

    def parse_connections(self):
        self.connections = []
        for entity in self.entities:
            if entity.raw_connections is not None:
                for side, connection_point in \
                        entity.raw_connections.items():
                    for color, connections in connection_point.items():
                        for connection in connections:
                            other = self.get_entity_by_id(
                                connection['entity_id'])
                            conn = Connection(
                                entity,
                                other,
                                side,
                                connection.get('circuit_id', '1'),
                                color=color)
                            if conn not in self.connections:
                                self.connections.append(conn)
        logger.debug(len(self.connections))
        for entity in self.entities:
            connections = entity.get_connections()



import json
import zlib
import base64
import math
from py_factorio_blueprints.exceptions import *


def _decode_0(string):
    try:
        data = json.loads(
            zlib.decompress(
                base64.b64decode(string[1:])).decode('UTF-8'))
    except (TypeError, base64.binascii.Error, zlib.error):
        raise InvalidExchangeString(
            "Could not decode exchange string")
    return data


def _encode_0(obj):
    temp = bytes(json.dumps(obj), 'UTF-8')
    temp2 = zlib.compress(temp)
    temp3 = base64.b64encode(temp2)
    return '0' + temp3.decode('UTF-8')


def decode(string):
    return _decode[string[0]](string)


def encode(obj):
    return _encode['latest'](obj)


_decode = {
    '0': _decode_0
}

_encode = {
    '0': _encode_0
}

_decode['latest'] = _decode['0']
_encode['latest'] = _encode['0']


class Color:
    def __init__(self, *args, **kwargs):
        props = ['r', 'g', 'b', 'a']
        for prop in props:
            setattr(self, prop, 1)

        if args:
            if len(args) > 4:
                raise TypeError(
                    f'Color() takes a maximum of 4 positional '
                    f'arguments ({len(args)} given)')
            for i, arg in enumerate(args):
                setattr(self, props[i], arg)
        if 'r' in kwargs:
            self.r = kwargs['r']
        if 'g' in kwargs:
            self.g = kwargs['g']
        if 'b' in kwargs:
            self.b = kwargs['b']
        if 'a' in kwargs:
            self.a = kwargs['a']

    def __repr__(self):
        return f"<Color (r:{self.r}, g:{self.g}, b:{self.b}, a:{self.a})>"

    def __iter__(self):
        yield from [self.r, self.g, self.b, self.a]

    def __eq__(self, other):
        if not isinstance(other, Color):
            return NotImplemented
        return (
            self.r == other.r and
            self.g == other.g and
            self.b == other.b and
            self.a == other.a)

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, value):
        self._r = max(min(1, value), 0)

    @property
    def g(self):
        return self._g

    @g.setter
    def g(self, value):
        self._g = max(min(1, value), 0)

    @property
    def b(self):
        return self._b

    @b.setter
    def b(self, value):
        self._b = max(min(1, value), 0)

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, value):
        self._a = max(min(1, value), 0)

    def to_json(self):
        obj = {
            'r': self.r,
            'g': self.g,
            'b': self.b,
            'a': self.a
        }
        return obj


class Direction(int):
    CLOCKWISE = 0
    COUNTER_CLOCKWISE = 1

    def __new__(cls, value):
        return super().__new__(cls, value % 8)

    def __add__(self, other):
        return Direction(
            super().__add__(other))

    def __sub__(self, other):
        return Direction(
            super().__sub__(other))

    def __floordiv__(self, other):
        return NotImplemented

    def __truediv__(self, other):
        return NotImplemented

    def __repr__(self):
        dirs = (
            "Up", "Up-Right", "Right", "Down-Right",
            "Down", "Down-Left", "Left", "Up-Left"
        )
        return f"<Direction ({dirs[self]})>"

    def __str__(self):
        return str(int(self))

    @classmethod
    def up(cls):
        return cls(0)

    @classmethod
    def right(cls):
        return cls(2)

    @classmethod
    def down(cls):
        return cls(4)

    @classmethod
    def left(cls):
        return cls(6)

    @property
    def is_up(self):
        return self == 0

    @property
    def is_right(self):
        return self == 2

    @property
    def is_down(self):
        return self == 4

    @property
    def is_left(self):
        return self == 6

    def rotate45(self, amount, direction=CLOCKWISE):
        """rotates by 45 degrees"""
        if direction != self.CLOCKWISE:
            amount = 8 - amount
        return Direction(self + amount)

    def rotate(self, amount, direction=CLOCKWISE):
        """rotates by 90 degrees"""
        amount *= 2
        if direction != self.CLOCKWISE:
            amount = 8 - amount
        return Direction(self + amount)

    @property
    def vector(self):
        if self == 0:
            return Vector(0, -1)
        elif self == 1:
            return Vector(1, -1)
        elif self == 2:
            return Vector(1, 0)
        elif self == 3:
            return Vector(1, 1)
        elif self == 4:
            return Vector(0, 1)
        elif self == 5:
            return Vector(-1, 1)
        elif self == 6:
            return Vector(-1, 0)
        elif self == 7:
            return Vector(-1, -1)
        return ValueError("Direction outside of 0-7 range")


class Condition:
    def __init__(self, *, first_signal=None, comparator=None,
                 constant=None, second_signal=None):
        if "name" in first_signal:
            first_signal = first_signal["name"]
        self.first = first_signal
        self.second = second_signal or constant
        self.comparator = comparator

    @property
    def first(self):
        return self.__first

    @first.setter
    def first(self, value):
        self.__first = NameStr(value)

    @property
    def second(self):
        return self.__second

    @second.setter
    def second(self, value):
        if isinstance(value, int):
            self.__second = value
        else:
            self.__second = NameStr(value)

    @property
    def comparator(self):
        return self.__comparator

    @comparator.setter
    def comparator(self, value):
        if value not in [">", "<", "=", ">=", "<=", "!="]:
            raise ValueError(value)
        self.__comparator = value

    def to_json(self):
        condition = {}
        if self.first is not None:
            condition["first_signal"] = {
                "name": self.first,
                "type": self.first.type}
        condition["comparator"] = self.comparator
        if self.second is not None:
            if isinstance(self.second, int):
                condition["constant"] = self.second
            else:
                condition["second_signal"] = {
                    "name": self.second,
                    "type": self.second.type}
        return condition


class NameStr(str):
    def __new__(cls, value):
        from py_factorio_blueprints.defaultentities import \
            defaultentities as entity_data
        if value is None:
            return None
        if not entity_data.get(value, None):
            raise UnknownEntity(value)
        return super().__new__(cls, value)

    @property
    def metadata(self):
        from py_factorio_blueprints.defaultentities import \
            defaultentities as entity_data
        return entity_data.get(self)

    @property
    def type(self):
        from py_factorio_blueprints.defaultentities import \
            defaultentities as entity_data
        return entity_data.get(self)["type"]


class SignalID:
    def __init__(self, data):
        self.name = NameStr(data['name'])

    @property
    def type(self):
        from py_factorio_blueprints.defaultentities import \
            defaultentities as entity_data
        return entity_data[self.name]['type']

    def __repr__(self):
        return f"<SignalID ({self.name}, type: {self.type})>"

    def to_json(self):
        return {'name': self.name,
                'type': self.type}


class Connection:
    def __init__(self, from_entity, to_entity,
                 from_side=1, to_side=1, color='green'):
        self.from_entity = from_entity
        self.to_entity = to_entity
        self.from_side = int(from_side)
        self.to_side = int(to_side)
        self.color = color

    def __repr__(self):
        return f"<Connection ({self.color} " \
               f"from:({self.from_entity}, {self.from_side}) " \
               f"to:({self.to_entity}, {self.to_side}))>"

    def attached_to(self, entity):
        if self.from_entity is entity or self.to_entity is entity:
            return True
        return False

    def flip(self):
        self.from_entity, self.to_entity = self.to_entity, self.from_entity
        self.from_side, self.to_side = self.to_side, self.from_side

    def orientate(self, entity):
        if self.to_entity == entity:
            self.flip()
        return self

    def flip_color(self):
        if self.color == 'green':
            self.color = 'red'
        else:
            self.color = 'green'

    def __eq__(self, other):
        if type(other) is not Connection:
            return NotImplemented
        if self.color != other.color:
            return False
        if self.from_entity == other.from_entity and \
                self.to_entity == other.to_entity and \
                self.from_side == other.from_side and \
                self.to_side == other.to_side:
            return True
        elif self.from_entity == other.to_entity and \
                self.to_entity == other.from_entity and \
                self.from_side == other.to_side and \
                self.to_side == other.from_side:
            return True
        return False


class Vector:
    def __new__(cls, *args, **kwargs):
        if args and kwargs:
            raise TypeError("Vector() can't set both args and kwargs")
        if not args and not kwargs:
            return None
        if args:
            if args[0] is None:
                return None
        return super().__new__(cls)

    def __init__(self, *args, **kwargs):
        if args:
            if len(args) == 1:
                if type(args[0]) == dict:
                    self.x = args[0]["x"]
                    self.y = args[0]["y"]
                elif type(args[0]) == tuple or \
                        type(args[0]) == Vector:
                    self.x, self.y = args[0]
                else:
                    raise ValueError(args)
            elif len(args) == 2:
                self.x = args[0]
                self.y = args[1]
            else:
                raise ValueError(args)
        else:
            self.x = kwargs['x']
            self.y = kwargs['y']

    def to_json(self):
        return {
            'x': self.x,
            'y': self.y
        }

    def __iter__(self):
        yield from [self.x, self.y]

    @property
    def xy(self):
        return Vector(self.x, self.y)

    @property
    def yx(self):
        return Vector(self.y, self.x)

    def __repr__(self):
        return f"<Vector ({self.x}, {self.y})>"

    def __add__(self, other):
        if type(other) is Vector:
            return Vector(self.x + other.x, self.y + other.y)
        elif type(other) is int or type(other) is float:
            return Vector(self.x + other, self.y + other)
        elif type(other) is tuple:
            x, y = other
            return Vector(self.x + x, self.y + y)
        return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if type(other) is Vector:
            return Vector(self.x - other.x, self.y - other.y)
        elif type(other) is int or type(other) is float:
            return Vector(self.x - other, self.y - other)
        elif type(other) is tuple:
            x, y = other
            return Vector(self.x - x, self.y - y)
        return NotImplemented

    def __rsub__(self, other):
        if type(other) is int or type(other) is float:
            return Vector(other - self.x, other - self.y)
        elif type(other) is tuple:
            x, y = other
            return Vector(x - self.x, y - self.y)
        return NotImplemented

    def __mul__(self, other):
        if type(other) is Vector:
            return Vector(self.x * other.x, self.y * other.y)
        elif type(other) is int or type(other) is float:
            return Vector(self.x * other, self.y * other)
        elif type(other) is tuple:
            x, y = other
            return Vector(self.x * x, self.y * y)
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if type(other) is Vector:
            return Vector(self.x / other.x, self.y / other.y)
        elif type(other) is int or type(other) is float:
            return Vector(self.x / other, self.y / other)
        elif type(other) is tuple:
            x, y = other
            return Vector(self.x / x, self.y / y)
        return NotImplemented

    def __rtruediv__(self, other):
        if type(other) is int or type(other) is float:
            return Vector(other / self.x, other / self.y)
        elif type(other) is tuple:
            x, y = other
            return Vector(x / self.x, y / self.y)
        return NotImplemented

    def __floordiv__(self, other):
        if type(other) is Vector:
            return Vector(self.x // other.x, self.y // other.y)
        elif type(other) is int or type(other) is float:
            return Vector(self.x // other, self.y // other)
        elif type(other) is tuple:
            x, y = other
            return Vector(self.x // x, self.y // y)
        return NotImplemented

    def __rfloordiv__(self, other):
        if type(other) is int or type(other) is float:
            return Vector(other // self.x, other // self.y)
        elif type(other) is tuple:
            x, y = other
            return Vector(x // self.x, y // self.y)
        return NotImplemented

    def __mod__(self, other):
        if type(other) is Vector:
            return Vector(self.x % other.x, self.y % other.y)
        elif type(other) is int or type(other) is float:
            return Vector(self.x % other, self.y % other)
        elif type(other) is tuple:
            x, y = other
            return Vector(self.x % x, self.y % y)
        return NotImplemented

    def __rmod__(self, other):
        if type(other) is int:
            return Vector(other % self.x, other % self.y)
        elif type(other) is tuple:
            x, y = other
            return Vector(x % self.x, y % self.y)

    def __eq__(self, other):
        if type(other) is tuple:
            x, y = other
            if self.x == x and self.y == y:
                return True
            else:
                return False
        elif type(other) is not Vector:
            return NotImplemented
        if self.x == other.x and self.y == other.y:
            return True
        return False

    def __lt__(self, other):
        if type(other) is tuple:
            other = Vector(other)
        elif type(other) is not Vector:
            return NotImplemented
        if self.x < other.x and self.y < other.y:
            return True
        return False

    def __gt__(self, other):
        if type(other) is tuple:
            other = Vector(other)
        elif type(other) is not Vector:
            return NotImplemented
        if self.x > other.x and self.y > other.y:
            return True
        return False

    def ceil(self):
        x = math.ceil(self.x)
        y = math.ceil(self.y)
        return Vector(x, y)

    def floor(self):
        x = math.floor(self.x)
        y = math.floor(self.y)
        return Vector(x, y)

    def round(self):
        x = math.floor(self.x + 0.5)
        y = math.floor(self.y + 0.5)
        return Vector(x, y)

    def copy(self):
        return Vector(self.x, self.y)

    def rotate(self, amount):
        amount %= 4
        v = self.copy()
        for _ in range(amount):
            v.x, v.y = -v.y, v.x
        return v


class TileName:
    class NameStr(str):
        @property
        def data(self):
            from py_factorio_blueprints import Blueprint
            return Blueprint.tile_prototypes[self]

    def __set_name__(self, owner, name):
        self.name = "__" + name

    def __set__(self, instance, value):
        if not getattr(instance, 'strict', True):
            setattr(instance, self.name, value)
            return

        from py_factorio_blueprints.blueprint import Blueprint

        if value not in Blueprint.tile_prototypes:
            raise UnknownTile(value)
        setattr(instance, self.name, value)

    def __get__(self, instance, owner):
        return TileName.NameStr(getattr(instance, self.name, ""))


class Tile:
    name = TileName()

    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop('name')
        self.position = Vector(kwargs.pop('position'))
        super().__init__(*args, **kwargs)

    def to_json(self):
        return {'name': self.name,
                'position': self.position.to_json()}

    def top_left(self):
        return self.position - Vector(0.5, 0.5)

    def bottom_right(self):
        return self.position + Vector(0.5, 0.5)

    def collides(self, position):
        return self.top_left < position < self.bottom_right

    def rotate(self, amount, around=None, direction='clockwise'):
        if around is None:
            around = Vector(0, 0)
        position = self.position - around
        amount %= 4
        if direction != 'clockwise':
            amount = 4 - amount

        def r(v):
            return Vector(-v.y, v.x)

        for _ in range(amount):
            position = r(position)
        self.position = position + around


def obj_set(obj, key, value):
    if value is None:
        return

    to_json = getattr(value, "to_json", None)
    if callable(to_json):
        value = to_json()
        if value is None:
            return

    obj[key] = value
    return obj


class ControlBehaviorMeta(type):
    def __new__(mcs, name, bases, attrs):
        control_behavior = attrs.pop('ControlBehavior', None)
        new_cls = super().__new__(mcs, name, bases, attrs)

        control_behaviors = [
            control_behavior,
            *[getattr(base, 'ControlBehavior', None)
              for base in bases]]
        control_behaviors = tuple(x for x in control_behaviors if x is not None)

        if control_behaviors:
            control_behavior = type(
                'ControlBehavior',
                control_behaviors,
                {})
            new_cls.ControlBehavior = control_behavior

        return new_cls



from py_factorio_blueprints.util import (
    NameStr, Vector, Color as ColorObj, Condition, obj_set, ControlBehaviorMeta)
from py_factorio_blueprints.exceptions import *


class Base(metaclass=ControlBehaviorMeta):
    pass


class BaseMixin(Base):
    NR_CONNECTIONS = 1
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            self.__entity = kwargs.pop('_entity')
            if not self.__entity:
                raise TypeError("Entity can't be NoneType")
            if args:
                raise TypeError(
                    f"{self.__class__}.__init__() "
                    f"({self.__entity.name}({self.__entity.name.data['type']})) "
                    f"takes no arguments. "
                    f"arguments provided: {args}")
            if kwargs:
                raise TypeError(
                    f"{self.__class__}.__init__() "
                    f"({self.__entity.name}({self.__entity.name.data['type']})) "
                    f"takes no keyword arguments. "
                    f"keyword arguments provided: {kwargs}")

        def to_json(self, obj):
            if not obj:
                return None
            return obj

    def __init__(self, *args, **kwargs):
        if args:
            raise TypeError(
                f"{self.__class__}.__init__() "
                f"({self.name}({self.name.data['type']})) "
                f"takes no arguments. "
                f"arguments provided: {args}")
        if kwargs:
            raise TypeError(
                f"{self.__class__}.__init__() "
                f"({self.name}({self.name.data['type']})) "
                f"takes no keyword arguments. "
                f"keyword arguments provided: {kwargs}")

    def to_json(self, obj):
        if hasattr(self, 'control_behavior'):
            obj_set(obj, 'control_behavior', self.control_behavior.to_json({}))
        return obj


class Operator(Base):
    def __init__(self, options):
        self.__options = options

    def __set_name__(self, owner, name):
        self.name = "__" + name

    def __set__(self, instance, value):
        if value not in [*self.__options, None]:
        # if value not in [">", "<", "=", "≥", "≤", "!=", None]:
            raise ValueError(value)
        setattr(instance, self.name, value)

    def __get__(self, instance, owner):
        return getattr(instance, self.name, None)


class SignalName(Base):
    class NameStr(str):
        @property
        def data(self):
            from py_factorio_blueprints.blueprint import Blueprint
            return Blueprint.signal_prototypes[self]

        def to_json(self):
            return {
                'name': self,
                'type': self.data['type']
            }

    def __set_name__(self, owner, name):
        self.name = "__" + name

    def __set__(self, instance, value):
        if not getattr(instance, 'strict', True):
            setattr(instance, self.name, value)
            return

        if type(value) is int:
            setattr(instance, self.name, value)
            return
        if type(value) is dict:
            value = value.get('name', None)

        from py_factorio_blueprints.blueprint import Blueprint

        if value and value not in Blueprint.signal_prototypes:
            raise UnknownSignal(value)
        setattr(instance, self.name, value)

    def __get__(self, instance, owner):
        value = getattr(instance, self.name, None)
        if value is None:
            return None
        if type(value) is int:
            return value
        return SignalName.NameStr(value)


class Accumulator(BaseMixin):
    class ControlBehavior:
        output_signal = SignalName()

        def __init__(self, *args, **kwargs):
            self.output_signal = kwargs.pop(
                'output_signal', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'output_signal', self.output_signal)
            return super().to_json(obj)


class Items(BaseMixin):
    class Item(Base):
        name = SignalName()

        def __init__(self, *, name, count):
            self.name = name
            self.count = count

    def __init__(self, *args, **kwargs):
        items = kwargs.pop('items', {})
        self.items = [
            self.Item(name=item, count=count)
            for item, count in items.items()]
        super().__init__(*args, *kwargs)

    def to_json(self, obj):
        if self.items:
            obj['items'] = {
                item.name: item.count
                for item in self.items}

        return super().to_json(obj)


class Filters(BaseMixin):
    class Filter(Base):
        signal = SignalName()

        def __init__(self, *args, **kwargs):
            self.signal = kwargs.pop(
                'signal', None)
            self.count = kwargs.pop(
                'count', None)
            self.index = kwargs.pop(
                'index', None)
            super().__init__(*args, *kwargs)

        def to_json(self):
            return {
                'signal': self.signal.to_json(),
                'count': self.count,
                'index': self.index
            }

    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            entity = kwargs.get('_entity')
            filters = kwargs.pop('filters', [])
            self.filters = [
                entity.Filter(**filter)
                for filter in filters]
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            if self.filters:
                obj['filters'] = [
                    filter.to_json()
                    for filter in self.filters]
            return super().to_json(obj)


class ConstantCombinator(Filters):
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            self.is_on = kwargs.pop('is_on', True)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            if not self.is_on:
                obj['is_on'] = False
            return super().to_json(obj)


class Combinator(BaseMixin):
    NR_CONNECTIONS = 2
    TYPE = ''
    OPERATOR = ''

    class ControlBehavior:
        first = SignalName()
        second = SignalName()
        output = SignalName()

        def __init__(self, *args, **kwargs):
            _entity = kwargs.get('_entity')
            self.__field = f"{_entity.TYPE}_conditions"
            conditions = kwargs.pop(self.__field, {})
            self.operator = conditions.pop(_entity.OPERATOR)

            def get_from(d, key):
                if _entity.TYPE == 'decider' and key == 'first':
                    value = d.pop("constant", None)
                else:
                    value = d.pop(f"{key}_constant", None)
                if value is None:
                    value = d.pop(f"{key}_signal", None)
                    if value is None:
                        return None
                    return value["name"]
                return value

            self.first = get_from(conditions, "first")
            self.second = get_from(conditions, "second")
            self.output = get_from(conditions, "output")
            if conditions:
                kwargs[self.__field] = conditions
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            def set_to(combinator, key, value):
                if type(value) is int:
                    if combinator == 'decider_conditions':
                        return "constant", value
                    else:
                        return f"{key}_constant", value
                else:
                    return f"{key}_signal", value
            fields = ['first', 'second', 'output']
            combinator_settings = {}
            combinator_settings[self.__entity.OPERATOR] = self.operator
            for field in fields:
                key, value = set_to(self.__field, field, getattr(self, field))
                obj_set(combinator_settings, key, value)
            if combinator_settings:
                if self.__field not in obj:
                    obj[self.__field] = {}
                obj[self.__field] = {**combinator_settings, **obj[self.__field]}
            return super().to_json(obj)


class Arithmetic(Combinator):
    TYPE = 'arithmetic'
    OPERATOR = 'operation'

    class ControlBehavior:
        operator = Operator(
            ['*', '/', '+', '-', '%', '^', '<<', '>>', 'AND', 'OR', 'XOR'])


class Decider(Combinator):
    TYPE = 'decider'
    OPERATOR = 'comparator'

    class ControlBehavior:
        operator = Operator([">", "<", "=", "≥", "≤", "!="])
        def __init__(self, *args, **kwargs):
            _entity = kwargs.get('_entity')
            self.__field = f"{_entity.TYPE}_conditions"
            conditions = kwargs.pop(self.__field)
            self.copy_count_from_input = conditions.pop(
                'copy_count_from_input', False)
            if conditions:
                kwargs[self.__field] = conditions
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            if not obj.get(self.__field, None):
                obj[self.__field] = {}
            obj[self.__field]['copy_count_from_input'] =\
                self.copy_count_from_input
            return super().to_json(obj)


class TransportBelt(BaseMixin):
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            self.circuit_enable_disable = kwargs.pop(
                'circuit_enable_disable', None)
            self.circuit_read_hand_contents = kwargs.pop(
                'circuit_read_hand_contents', None)
            self.circuit_contents_read_mode = kwargs.pop(
                'circuit_contents_read_mode', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj['circuit_read_hand_contents'] = self.circuit_read_hand_contents
            obj['circuit_enable_disable'] = self.circuit_enable_disable
            obj['circuit_contents_read_mode'] = self.circuit_contents_read_mode
            return super().to_json(obj)


class TrainSignal(BaseMixin):
    class ControlBehavior:
        red_output_signal = SignalName()
        orange_output_signal = SignalName()
        green_output_signal = SignalName()

        def __init__(self, *args, **kwargs):
            self.circuit_read_signal = kwargs.pop(
                'circuit_read_signal', None)
            self.red_output_signal = kwargs.pop(
                'red_output_signal', None)
            self.orange_output_signal = kwargs.pop(
                'orange_output_signal', None)
            self.green_output_signal = kwargs.pop(
                'green_output_signal', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'circuit_read_signal', self.circuit_read_signal)
            obj_set(obj, 'red_output_signal', self.red_output_signal)
            obj_set(obj, 'orange_output_signal', self.orange_output_signal)
            obj_set(obj, 'green_output_signal', self.green_output_signal)
            return super().to_json(obj)


class ChainSignal(TrainSignal):
    class ControlBehavior:
        blue_output_signal = SignalName()

        def __init__(self, *args, **kwargs):
            self.blue_output_signal = kwargs.pop(
                'blue_output_signal', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'blue_output_signal', self.blue_output_signal)
            return super().to_json(obj)


class RailSignal(TrainSignal):
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            self.circuit_close_signal = kwargs.pop(
                'circuit_close_signal', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'circuit_close_signal', self.circuit_close_signal)
            return super().to_json(obj)


class Orientation(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def rotate(self, amount, **kwargs):
        self.orientation = self.orientation + (0.25 * amount) % 1
        sup = super()
        if hasattr(sup, 'rotate'):
            sup.rotate(amount, **kwargs)


class Train(Orientation):
    def __init__(self, *args, orientation=None, **kwargs):
        self.orientation = orientation
        super().__init__(*args, **kwargs)

    @property
    def orientation(self):
        return self.__orientation

    @orientation.setter
    def orientation(self, value):
        if value is None:
            self.__orientation = None
        else:
            self.__orientation = value % 1.0

    def to_json(self, obj):
        obj_set(obj, 'orientation', self.orientation)
        return super().to_json(obj)


class Cargo(Train):
    def __init__(self, *args, inventory=None, **kwargs):
        self.inventory = inventory
        super().__init__(*args, **kwargs)

    @property
    def inventory_filters(self):
        return self.__inventory_filters

    @property
    def inventory_bar(self):
        return self.__inventory_bar

    @inventory_bar.setter
    def inventory_bar(self, value):
        if value is None:
            self.__inventory_bar = None
        else:
            self.__inventory_bar = int(value)

    @property
    def inventory(self):
        if self.__inventory_bar is None and not len(self.__inventory_filters):
            return None
        inventory = {}
        obj_set(inventory, 'bar', self.__inventory_bar)
        if len(self.__inventory_filters):
            inventory["filters"] = [
                {"index": index, "name": name}
                for index, name in self.__inventory_filters.items()]
        return inventory

    @inventory.setter
    def inventory(self, value):
        self.__inventory_filters = []
        if value is None:
            self.__inventory_bar = None
            return
        self.__inventory_bar = value.get("bar", None)
        self.__inventory_filters = {}
        for f in value.get("filters", []):
            self.__inventory_filters[f["index"]] = NameStr(f["name"])

    def to_json(self, obj):
        obj_set(obj, 'inventory', self.inventory)
        return super().to_json(obj)


class ColorField(Base):
    def __set_name__(self, owner, name):
        self.__name = f"__{name}"

    def __set__(self, instance, value):
        if value is None:
            setattr(instance, self.__name, None)
        else:
            setattr(instance, self.__name, ColorObj(**value))

    def __get__(self, instance, owner):
        return getattr(instance, self.__name, None)

class Color(BaseMixin):
    color = ColorField()

    def __init__(self, *args, **kwargs):
        self.color = kwargs.pop('color', None)
        super().__init__(*args, **kwargs)

    def to_json(self, obj):
        obj_set(obj, 'color', self.color)
        return super().to_json(obj)


class Container(BaseMixin):
    def __init__(self, *args, bar=None, **kwargs):
        self.bar = bar
        super().__init__(*args, **kwargs)

    @property
    def bar(self):
        return self.__bar

    @bar.setter
    def bar(self, value):
        if value is None:
            self.__bar = None
        else:
            self.__bar = int(value)

    def to_json(self, obj):
        obj_set(obj, 'bar', self.bar)
        return super().to_json(obj)


class CircuitCondition(BaseMixin):
    class ControlBehavior:
        first = SignalName()
        comparator = Operator([">", "<", "=", "≥", "≤", "!="])
        second = SignalName()

        def __init__(self, *args,  circuit_condition=None, **kwargs):
            if circuit_condition is None:
                circuit_condition = {}
            else:
                self.first = circuit_condition.get(
                    'first_signal', None)
                self.second = circuit_condition.get(
                    'second_signal',
                    circuit_condition.get('constant', None))
                self.comparator = circuit_condition.get('comparator', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            circuit_condition = {}
            obj_set(circuit_condition, 'comparator', self.comparator)
            if self.first is not None:
                if type(self.first) is int:
                    circuit_condition['first_constant'] = self.first
                else:
                    circuit_condition['first_signal'] = self.first.to_json()
            if self.second is not None:
                if type(self.second) is int:
                    circuit_condition['constant'] = self.second
                else:
                    circuit_condition['second_signal'] = self.second.to_json()
            if circuit_condition:
                obj['circuit_condition'] = circuit_condition
            return super().to_json(obj)


class Lamp(CircuitCondition):
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            self.use_colors = kwargs.pop(
                'use_colors', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'use_colors', self.use_colors)
            return super().to_json(obj)


class MiningDrill(CircuitCondition):
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            self.circuit_enable_disable = kwargs.pop(
                'circuit_enable_disable', True)
            self.circuit_read_resources = kwargs.pop(
                'circuit_read_resources', True)
            self.circuit_resource_read_mode = kwargs.pop(
                'circuit_resource_read_mode', 0)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj['circuit_enable_disable'] = self.circuit_enable_disable
            obj['circuit_read_resources'] = self.circuit_read_resources
            obj['circuit_resource_read_mode'] = self.circuit_resource_read_mode
            return super().to_json(obj)


class Inserter(BaseMixin):
    class ControlBehavior:
        stack_control_input_signal = SignalName()

        def __init__(self, *args, **kwargs):
            self.circuit_mode_of_operation = kwargs.pop(
                'circuit_mode_of_operation', None)
            self.circuit_set_stack_size = kwargs.pop(
                'circuit_set_stack_size', None)
            self.stack_control_input_signal = kwargs.pop(
                'stack_control_input_signal', None)
            self.circuit_hand_read_mode = kwargs.pop(
                'circuit_hand_read_mode', None)
            self.circuit_read_hand_contents = kwargs.pop(
                'circuit_read_hand_contents', None)

        def to_json(self, obj):
            obj_set(obj, 'circuit_mode_of_operation',
                    self.circuit_mode_of_operation)
            obj_set(obj, 'circuit_set_stack_size', self.circuit_set_stack_size)
            obj_set(obj, 'stack_control_input_signal',
                    self.stack_control_input_signal)
            obj_set(obj, 'circuit_hand_read_mode', self.circuit_hand_read_mode)
            obj_set(obj, 'circuit_read_hand_contents',
                    self.circuit_read_hand_contents)

            return super().to_json(obj)

    def __init__(self, *args, override_stack_size=None, drop_position=None,
                 pickup_position=None, **kwargs):
        self.override_stack_size = override_stack_size
        self.pickup_position = pickup_position
        self.drop_position = drop_position
        super().__init__(*args, **kwargs)

    @property
    def pickup_position(self):
        return self.__pickup_position

    @pickup_position.setter
    def pickup_position(self, value):
        self.__pickup_position = Vector(value)

    @property
    def drop_position(self):
        return self.__drop_position

    @drop_position.setter
    def drop_position(self, value):
        self.__drop_position = Vector(value)

    @property
    def override_stack_size(self):
        return self.__override_stack_size

    @override_stack_size.setter
    def override_stack_size(self, value):
        if value is not None:
            value = int(value)
        self.__override_stack_size = value

    def to_json(self, obj):
        obj_set(obj, 'override_stack_size', self.override_stack_size)
        obj_set(obj, 'pickup_position', self.pickup_position)
        obj_set(obj, 'drop_position', self.drop_position)
        return super().to_json(obj)


class FilterInserter(Inserter):
    class Filter(Base):
        name = SignalName()

        def __init__(self, *args, **kwargs):
            self.name = kwargs.pop(
                'name', None)
            self.index = kwargs.pop(
                'index', None)
            super().__init__(*args, *kwargs)

        def to_json(self):
            return {'index': self.index, 'name': self.name}

    def __init__(self, *args, filters=None, filter_mode=None, **kwargs):
        if filters is None:
            filters = []
        self.filters = [
            self.Filter(**filter)
            for filter in filters]
        self.filter_mode = filter_mode
        super().__init__(*args, **kwargs)

    @property
    def filter_mode(self):
        return self.__filter_mode

    @filter_mode.setter
    def filter_mode(self, value):
        if value not in [None, "whitelist", "blacklist"]:
            raise ValueError(value)
        self.__filter_mode = value

    def to_json(self, obj):
        obj_set(obj, "filter_mode", self.filter_mode)
        if self.filters:
            obj["filters"] = [
                filter.to_json()
                for filter in self.filters]
        return super().to_json(obj)


class InfinityContainer(BaseMixin):
    def __init__(self, *args, infinity_settings=None, **kwargs):
        self.infinity_settings = infinity_settings
        super().__init__(*args, **kwargs)

    @property
    def infinity_settings(self):
        return {
            "remove_unfiltered_items":
                self.infinity_settings_remove_unfiltered_items,
            "filters": self.infinity_settings_filters
        }

    @infinity_settings.setter
    def infinity_settings(self, value):
        self.infinity_settings_remove_unfiltered_items =\
            value["remove_unfiltered_items"]
        self.infinity_settings_filters = value["filters"]

    @property
    def infinity_settings_remove_unfiltered_items(self):
        return self.__infinity_settings_remove_unfiltered_items

    @infinity_settings_remove_unfiltered_items.setter
    def infinity_settings_remove_unfiltered_items(self, value):
        self.__infinity_settings_remove_unfiltered_items = bool(value)

    @property
    def infinity_settings_filters(self):
        return self.__infinity_settings_filters

    @infinity_settings_filters.setter
    def infinity_settings_filters(self, values):
        self.__infinity_settings_filters = {}
        if values is None:
            values = []
        for value in values:
            if value["mode"] not in ["at-least", "at-most", "exactly"]:
                raise ValueError(value["mode"])
            self.__infinity_settings_filters[value["index"]] = {
                "count": value["count"],
                "mode": value["mode"],
                "name": NameStr(value["name"])
            }

    def to_json(self, obj):
        if self.infinity_settings:
            obj["infinity_settings"] = {
                "remove_unfiltered_items":
                    self.infinity_settings_remove_unfiltered_items,
                "filters": [
                    {
                        "index": index,
                        "count": value["count"],
                        "mode": value["mode"],
                        "name": value["name"]
                    }
                    for index, value in
                    self.infinity_settings_filters.items()
                ]
            }
        return super().to_json(obj)


class RecipeName(Base):
    class NameStr(str):
        @property
        def data(self):
            from py_factorio_blueprints.blueprint import Blueprint
            return Blueprint.recipe_prototypes[self]

    def __set_name__(self, owner, name):
        self.name = "__" + name

    def __set__(self, instance, value):
        if not getattr(instance, 'strict', True):
            setattr(instance, self.name, value)
            return

        from py_factorio_blueprints.blueprint import Blueprint

        if value is not None and value not in Blueprint.recipe_prototypes:
            raise UnknownRecipe(value)
        setattr(instance, self.name, value)

    def __get__(self, instance, owner):
        value = getattr(instance, self.name, None)
        if value is None:
            return None
        return RecipeName.NameStr(value)


class Recipe(BaseMixin):
    recipe = RecipeName()

    def __init__(self, *args, recipe=None, **kwargs):
        self.recipe = recipe
        super().__init__(*args, **kwargs)

    def to_json(self, obj):
        obj_set(obj, 'recipe', self.recipe)
        return super().to_json(obj)


class Requester(BaseMixin):
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            self.circuit_mode_of_operation = kwargs.pop(
                'circuit_mode_of_operation', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'circuit_mode_of_operation',
                    self.circuit_mode_of_operation)
            return super().to_json(obj)

    def __init__(self, *args,
                 request_from_buffers=False, request_filters=None, **kwargs):
        self.request_from_buffers = request_from_buffers
        self.request_filters = request_filters
        super().__init__(*args, **kwargs)

    @property
    def request_from_buffers(self):
        return self.__request_from_buffers

    @request_from_buffers.setter
    def request_from_buffers(self, value):
        self.__request_from_buffers = value

    @property
    def request_filters(self):
        return self.__request_filters

    @request_filters.setter
    def request_filters(self, values):
        if values is None:
            values = []
        self.__request_filters = {}
        for value in values:
            self.__request_filters[value["index"]] = {
                "name": NameStr(value["name"]),
                "count": int(value["count"])
            }

    def to_json(self, obj):
        if self.request_from_buffers:
            obj["request_from_buffers"] = self.request_from_buffers
        if self.request_filters:
            obj["request_filters"] = [
                {
                    "index": index,
                    "name": f["name"],
                    "count": f["count"]
                }
                for index, f in self.request_filters.items()
            ]
        return super().to_json(obj)


class Roboport(BaseMixin):
    class ControlBehavior:
        available_logistic_output_signal = SignalName()
        total_logistic_output_signal = SignalName()
        available_construction_output_signal = SignalName()
        total_construction_output_signal = SignalName()

        def __init__(self, *args, **kwargs):
            self.read_logistics = kwargs.pop(
                'read_logistics', None)
            self.read_robot_stats = kwargs.pop(
                'read_robot_stats', None)
            self.available_construction_output_signal = kwargs.pop(
                'available_construction_output_signal', None)
            self.total_construction_output_signal = kwargs.pop(
                'total_construction_output_signal', None)
            self.available_logistic_output_signal = kwargs.pop(
                'available_logistic_output_signal', None)
            self.total_logistic_output_signal = kwargs.pop(
                'total_logistic_output_signal', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'read_logistics', self.read_logistics)
            obj_set(obj, 'read_robot_stats', self.read_robot_stats)
            fields = [
                'available_construction_output_signal',
                'total_construction_output_signal',
                'available_logistic_output_signal',
                'total_logistic_output_signal'
            ]
            for field in fields:
                if getattr(self, field):
                    obj[field] = getattr(self, field).to_json()
            return super().to_json(obj)


class Rotatable(BaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def rotate(self, amount, **kwargs):
        self.direction = self.direction.rotate(amount)
        sup = super()
        if hasattr(sup, "rotate"):
            sup.rotate(amount, **kwargs)

    def to_json(self, obj):
        if not self.direction.is_up:
            obj['direction'] = self.direction
        return super().to_json(obj)


class Silo(BaseMixin):
    def __init__(self, *args, auto_launch=None, **kwargs):
        self.auto_launch = auto_launch
        super().__init__(*args, **kwargs)

    @property
    def auto_launch(self):
        return self.__auto_launch

    @auto_launch.setter
    def auto_launch(self, value):
        if value not in [None, True, False]:
            raise ValueError
        self.__auto_launch = value

    def to_json(self, obj):
        obj_set(obj, 'auto_launch', self.auto_launch)
        return super().to_json(obj)


class Speaker(CircuitCondition):
    class ControlBehavior:
        def __init__(self, *args, **kwargs):
            entity = kwargs.get('_entity')
            self.circuit_parameters = entity.CircuitParameters(
                **kwargs.pop('circuit_parameters', {}))
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj['circuit_parameters'] = self.circuit_parameters.to_json()
            return super().to_json(obj)

    class CircuitParameters:
        def __init__(self, *args, **kwargs):
            self.signal_value_is_pitch = kwargs.pop(
                'signal_value_is_pitch', None)
            self.instrument_id = kwargs.pop(
                'instrument_id', None)
            self.note_id = kwargs.pop(
                'note_id', None)
            super().__init__(*args, **kwargs)

        def to_json(self):
            return {
                'instrument_id': self.instrument_id,
                'note_id': self.note_id,
                'signal_value_is_pitch': self.signal_value_is_pitch}

    def __init__(self, *args, parameters=None, alert_parameters=None, **kwargs):
        self.parameters = parameters
        self.alert_parameters = alert_parameters
        super().__init__(*args, **kwargs)

    @property
    def parameters(self):
        return self.__parameters

    @parameters.setter
    def parameters(self, value):
        if value is None:
            self.__parameters = None
            return
        self.__parameters = {
            "playback_volume": float(value["playback_volume"]),
            "playback_globally": bool(value["playback_globally"]),
            "allow_polyphony": bool(value["allow_polyphony"])
        }

    @property
    def alert_parameters(self):
        return self.__alert_parameters

    @alert_parameters.setter
    def alert_parameters(self, value):
        if value is None:
            self.__alert_parameters = None
            return
        self.__alert_parameters = {
            "show_alert": bool(value["show_alert"]),
            "show_on_map": bool(value["show_on_map"]),
            "icon_signal_id": NameStr(
                value.get("icon_signal_id", {}).get("name", None)),
            "alert_message": value["alert_message"]
        }

    def to_json(self, obj):
        obj_set(obj, 'parameters', self.parameters)
        if self.alert_parameters is not None:
            sobj = {
                "show_alert": self.alert_parameters["show_alert"],
                "show_on_map": self.alert_parameters["show_on_map"],
                "alert_message": self.alert_parameters["alert_message"]}
            if self.alert_parameters["icon_signal_id"] is not None:
                sobj["icon_signal_id"] = {
                    "name": self.alert_parameters["icon_signal_id"],
                    "type": self.alert_parameters["icon_signal_id"].type}
            obj["alert_parameters"] = sobj
        return super().to_json(obj)


class Splitter(BaseMixin):
    def __init__(self, *args,
                 input_priority=None, output_priority=None, filter=None,
                 **kwargs):
        self.input_priority = input_priority
        self.output_priority = output_priority
        self.filter = filter
        super().__init__(*args, **kwargs)

    @property
    def filter(self):
        return self.__filter

    @filter.setter
    def filter(self, value):
        self.__filter = NameStr(value)

    @property
    def input_priority(self):
        return self.__input_priority

    @input_priority.setter
    def input_priority(self, value):
        if value not in [None, "right", "left"]:
            raise ValueError(value)
        self.__input_priority = value

    @property
    def output_priority(self):
        return self.__output_priority

    @output_priority.setter
    def output_priority(self, value):
        if value not in [None, "right", "left"]:
            raise ValueError(value)
        self.__output_priority = value

    def to_json(self, obj):
        obj_set(obj, 'input_priority', self.input_priority)
        obj_set(obj, 'output_priority', self.output_priority)
        obj_set(obj, 'filter', self.filter)
        return super().to_json(obj)


class Station(CircuitCondition):
    class ControlBehavior:
        train_stopped_signal = SignalName()

        def __init__(self, *args, **kwargs):
            self.circuit_enable_disable = kwargs.pop(
                'circuit_enable_disable', None)
            self.read_from_train = kwargs.pop(
                'read_from_train', None)
            self.read_stopped_train = kwargs.pop(
                'read_stopped_train', None)
            self.train_stopped_signal = kwargs.pop(
                'train_stopped_signal', None)
            super().__init__(*args, **kwargs)

        def to_json(self, obj):
            obj_set(obj, 'circuit_enable_disable', self.circuit_enable_disable)
            obj_set(obj, 'read_from_train', self.read_from_train)
            obj_set(obj, 'read_stopped_train', self.read_stopped_train)
            if self.train_stopped_signal:
                obj_set(obj, 'train_stopped_signal',
                        self.train_stopped_signal.to_json())
            return super().to_json(obj)

    def __init__(self, *args, station=None, **kwargs):
        self.station = station
        super().__init__(*args, **kwargs)

    def to_json(self, obj):
        obj_set(obj, 'station', self.station)
        return super().to_json(obj)


class Underground(BaseMixin):
    def __init__(self, *args, type, **kwargs):
        self.type = type
        super().__init__(*args, **kwargs)

    @property
    def type(self):
        return self.__type

    @type.setter
    def type(self, value):
        if value not in ["input", "output"]:
            raise ValueError(value)
        self.__type = value

    def to_json(self, obj):
        obj_set(obj, 'type', self.type)
        return super().to_json(obj)


class Loader(Underground):
    pass


class Variation(BaseMixin):
    def __init__(self, *args, variation=None, **kwargs):
        self.variation = variation
        super().__init__(*args, **kwargs)

    @property
    def variation(self):
        return self.__variation

    @variation.setter
    def variation(self, value):
        if value is not None and \
                value < 0 or value > 255:
            raise ValueError(value)
        self.__variation = value

    def to_json(self, obj):
        obj_set(obj, 'variation', self.variation)
        return super().to_json(obj)



from py_factorio_blueprints.entity_mixins import *
fluid = 'fluid'
item = 'item'
virtual = 'virtual'
tile = 'tile'
recipe = 'recipe'


defaultentities = {   # ADD MORE (vanilla) AS YOU PLEASE (or modded if it's just for you)!
                      # Somebody will probably automate the gathering of this data soon...

    'programmable-speaker': {
        'mixins': [Speaker, CircuitCondition],
        'type': item,
        'width': 1,
        'height': 1,


        'parameters': True,
        'alertParameters': True
    },

    'heat-exchanger': {
        'mixins': [Rotatable],
        'type': item,
        'width': 3,
        'height': 2
    },

    'heat-pipe': {
        'type': item,
        'width': 1,
        'height': 1
    },

    'nuclear-reactor': {
        'type': item,
        'width': 5,
        'height': 5
    },

    'centrifuge': {
        'mixins': [Recipe, Items],
        'type': item,
        'width': 3,
        'height': 3
    },

    'steam-turbine': {
        'mixins': [Rotatable],
        'type': item,
        'width': 3,
        'height': 5
    },

    'tank': {
        'type': item
    },
    'car': {
        'type': item
    },
    'cargo-wagon': {
        'mixins': [Cargo],
        'type': item,
        'inventorySize': 40,
        'width': 1,
        'height': 1
    },
    'fluid-wagon': {
        'type': item,
    },
    'locomotive': {
        'type': item
    },

    'light-armor': {
        'type': item
    },
    'heavy-armor': {
        'type': item
    },
    'modular-armor': {
        'type': item
    },
    'grenade': {
        'type': item
    },
    'cluster-grenade': {
        'type': item
    },
    'flamethrower': {
        'type': item
    },
    'flamethrower-ammo': {
        'type': item
    },
    'rocket-launcher': {
        'type': item
    },
    'rocket': {
        'type': item
    },
    'explosive-rocket': {
        'type': item
    },
    'atomic-bomb': {
        'type': item
    },
    'combat-shotgun': {
        'type': item
    },
    'shotgun': {
        'type': item
    },
    'shotgun-shell': {
        'type': item
    },
    'piercing-shotgun-shell': {
        'type': item
    },
    'submachine-gun': {
        'type': item
    },
    'pistol': {
        'type': item
    },
    'firearm-magazine': {
        'type': item
    },
    'piercing-rounds-magazine': {
        'type': item
    },
    'uranium-rounds-magazine': {
        'type': item
    },
    'cannon-shell': {
        'type': item
    },
    'explosive-cannon-shell': {
        'type': item
    },
    'uranium-cannon-shell': {
        'type': item
    },
    'explosive-uranium-cannon-shell': {
        'type': item
    },

    'power-armor': {
        'type': item
    },
    'power-armor-mk2': {
        'type': item
    },
    'energy-shield-equipment': {
        'type': item
    },
    'energy-shield-mk2-equipment': {
        'type': item
    },
    'solar-panel-equipment': {
        'type': item
    },
    'fusion-reactor-equipment': {
        'type': item
    },
    'battery-equipment': {
        'type': item
    },
    'battery-mk2-equipment': {
        'type': item
    },
    'personal-laser-defense-equipment': {
        'type': item
    },
    'discharge-defense-equipment': {
        'type': item
    },
    'exoskeleton-equipment': {
        'type': item
    },
    'personal-roboport-equipment': {
        'type': item
    },
    'personal-roboport-mk2-equipment': {
        'type': item
    },
    'night-vision-equipment': {
        'type': item
    },

    'discharge-defense-remote': {
        'type': item
    },
    'destroyer-capsule': {
        'type': item
    },
    'distractor-capsule': {
        'type': item
    },
    'defender-capsule': {
        'type': item
    },
    'slowdown-capsule': {
        'type': item
    },
    'poison-capsule': {
        'type': item
    },



    'stone': {
        'type': item
    },

    'solid-fuel': {
        'type': item
    },

    'stone-brick': {
        'type': item
    },

    'stone-path': {
        'type': tile
    },
    'landfill': {
        'type': item
    },
    'concrete': {
        'type': tile
    },
    'hazard-concrete': {
        'type': item
    },
    'hazard-concrete-left': {
        'type': tile
    },
    'hazard-concrete-right': {
        'type': tile
    },
    'refined-concrete': {
        'type': tile
    },
    'refined-hazard-concrete': {
        'type': item
    },
    'refined-hazard-concrete-left': {
        'type': tile
    },
    'refined-hazard-concrete-right': {
        'type': tile
    },

    'iron-axe': {
        'type': item
    },
    'steel-axe': {
        'type': item
    },
    'repair-pack': {
        'type': item
    },
    'blueprint': {
        'type': item
    },
    'deconstruction-planner': {
        'type': item
    },
    'blueprint-book': {
        'type': item
    },

    'copper-cable': {
        'type': item
    },
    'red-wire': {
        'type': item
    },
    'green-wire': {
        'type': item
    },

    'beacon': {
        'type': item,
        'width': 3,
        'height': 3,

        'modules': 2
    },
    'small-electric-pole': {
        'type': item,
        'width': 1,
        'height': 1
    },
    'medium-electric-pole': {
        'type': item,
        'width': 1,
        'height': 1
    },
    'substation': {
        'type': item,
        'width': 2,
        'height': 2
    },
    'big-electric-pole': {
        'type': item,
        'width': 2,
        'height': 2
    },
    'offshore-pump': {
        'mixins': [Rotatable],
        'type': item,
        'width': 2,
        'height': 2
    },
    'small-lamp': {
        'type': item,
        'width': 1,
        'height': 1
    },
    'solar-panel': {
        'type': item,
        'width': 3,
        'height': 3
    },
    'arithmetic-combinator': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 2
    },
    'decider-combinator': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 2
    },
    'constant-combinator': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },

    'splitter': {   # Default position is facing north, 2 wide and 1 high for all splitters.
        'mixins': [Rotatable, Splitter],
        'type': item,
        'width': 2,
        'height': 1
    },
    'fast-splitter': {
        'mixins': [Rotatable, Splitter],
        'type': item,
        'width': 2,
        'height': 1
    },
    'express-splitter': {
        'mixins': [Rotatable, Splitter],
        'type': item,
        'width': 2,
        'height': 1
    },
    'transport-belt': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    'fast-transport-belt': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    'express-transport-belt': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    'underground-belt': {
        'mixins': [Rotatable, Underground],
        'type': item,
        'width': 1,
        'height': 1,
        'directionType': True
    },
    'fast-underground-belt': {
        'mixins': [Rotatable, Underground],
        'type': item,
        'width': 1,
        'height': 1,
        'directionType': True
    },
    'express-underground-belt': {
        'mixins': [Rotatable, Underground],
        'type': item,
        'width': 1,
        'height': 1,
        'directionType': True
    },
    'assembling-machine-1': {
        'mixins': [Recipe, Items],
        'type': item,
        'width': 3,
        'height': 3,
        'recipe': True
    },
    'assembling-machine-2': {
        'mixins': [Recipe, Items],
        'type': item,
        'width': 3,
        'height': 3,

        'recipe': True,
        'modules': 2
    },
    'assembling-machine-3': {
        'mixins': [Recipe, Items],
        'type': item,
        'width': 3,
        'height': 3,

        'recipe': True,
        'modules': 4
    },
    'wooden-chest': {
        'mixins': [Container],
        'type': item,
        'width': 1,
        'height': 1,

        'inventorySize': 16
    },
    'iron-chest': {
        'mixins': [Container],
        'type': item,
        'width': 1,
        'height': 1,

        'inventorySize': 32
    },
    'steel-chest': {
        'mixins': [Container],
        'type': item,
        'width': 1,
        'height': 1,
        'inventorySize': 48
    },
    'infinity-chest': {
        'mixins': [InfinityContainer],
        'type': item,
        'width': 1,
        'height': 1,
        'inventorySize': 48
    },
    'logistic-chest-passive-provider': {
        'mixins': [Container],
        'type': item,
        'width': 1,
        'height': 1,
        'inventorySize': 48
    },
    'logistic-chest-active-provider': {
        'mixins': [Container],
        'type': item,
        'width': 1,
        'height': 1,
        'inventorySize': 48
    },
    'logistic-chest-storage': {
        'mixins': [Container],
        'type': item,
        'width': 1,
        'height': 1,
        'inventorySize': 48
    },
    'logistic-chest-requester': {
        'mixins': [Container, Requester],
        'type': item,
        'width': 1,
        'height': 1,
        'inventorySize': 48
    },
    'logistic-chest-buffer': {
        'mixins': [Container, Requester],
        'type': item,
        'width': 1,
        'height': 1,
        'inventorySize': 48
    },
    'storage-tank': {
        'mixins': [Rotatable],
        'type': item,
        'width': 3,
        'height': 3
    },
    'burner-inserter': {
        'mixins': [Rotatable, Inserter],
        'type': item,
        'width': 1,
        'height': 1
    },
    'inserter': {
        'mixins': [Rotatable, Inserter],
        'type': item,
        'width': 1,
        'height': 1
    },
    'long-handed-inserter': {
        'mixins': [Rotatable, Inserter],
        'type': item,
        'width': 1,
        'height': 1
    },
    'fast-inserter': {
        'mixins': [Rotatable, Inserter],
        'type': item,
        'width': 1,
        'height': 1
    },
    'filter-inserter': {
        'mixins': [Rotatable, FilterInserter],
        'type': item,
        'width': 1,
        'height': 1,
        'filterAmount': False
    },
    'stack-inserter': {
        'mixins': [Rotatable, Inserter],
        'type': item,
        'width': 1,
        'height': 1,
    },
    'stack-filter-inserter': {
        'mixins': [Rotatable, FilterInserter],
        'type': item,
        'width': 1,
        'height': 1,
        'filterAmount': False
    },
    'gate': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    'stone-wall': {
        'type': item,
        'width': 1,
        'height': 1
    },
    'radar': {
        'type': item,
        'width': 3,
        'height': 3
    },
    'rail': {
        'type': item
    },
    'straight-rail': {
        'mixins': [Rotatable],
        'type': item,
        'width': 2,
        'height': 2
    },
    'curved-rail': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    # Lets figure out curved rails later. (1 curved rail deconstructs to 4 straight rails)
    'land-mine': {
        'type': item,
        'width': 1,
        'height': 1
    },
    'train-stop': {  # pretty sure this is a 1.2x1.2 centered in a 2x2 square.
        'mixins': [Rotatable],
        'type': item,
        'width': 2,
        'height': 2
    },
    'rail-signal': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    'rail-chain-signal': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    'lab': {
        'type': item,
        'width': 3,
        'height': 3,

        'modules': 2
    },
    'rocket-silo': {
        'mixins': [Recipe, Items, Silo],
        'type': item,
        'width': 9,
        'height': 10,  # unsure about these values, got them from code only (never counted it in game, but 10 sounds right.)

        'modules': 4
    },
    'chemical-plant': {
        'mixins': [Rotatable, Recipe, Items],
        'type': item,
        'width': 3,
        'height': 3,

        'modules': 3
    },
    'oil-refinery': {
        'mixins': [Rotatable, Recipe, Items],
        'type': item,
        'width': 5,
        'height': 5,

        'modules': 3
    },
    'stone-furnace': {
        'type': item,
        'width': 2,
        'height': 2
    },
    'steel-furnace': {
        'type': item,
        'width': 2,
        'height': 2
    },
    'electric-furnace': {
        'type': item,
        'width': 3,
        'height': 3,

        'modules': 2
    },
    'pumpjack': {
        'mixins': [Rotatable],
        'type': item,
        'width': 3,
        'height': 3
    },
    'burner-mining-drill': {
        'mixins': [Rotatable],
        'type': item,
        'width': 2,
        'height': 2
    },

    'electric-mining-drill': {
        'mixins': [Rotatable],
        'type': item,
        'width': 3,
        'height': 3,

        'modules': 3
    },
    'pump': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 2
    },
    'pipe': {
        'type': item,
        'width': 1,
        'height': 1
    },
    'pipe-to-ground': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },

    'electronic-circuit': {
        'type': item
    },
    'advanced-circuit': {
        'type': item
    },


    'boiler': {
        'mixins': [Rotatable],
        'type': item,
        'width': 1,
        'height': 1
    },
    'steam-engine': {
        'mixins': [Rotatable],
        'type': item,
        'width': 5,
        'height': 3
    },
    'accumulator': {
        'type': item,
        'width': 2,
        'height': 2
    },

    'roboport': {
        'type': item,
        'width': 4,
        'height': 4
    },
    'construction-robot': {
        'type': item
    },
    'logistic-robot': {
        'type': item
    },
    'power-switch': {
        'type': item,
        'width': 3,
        'height': 3
    },

    'gun-turret': {
        'mixins': [Rotatable],
        'type': item,
        'width': 2,
        'height': 2
    },
    'laser-turret': {
        'mixins': [Rotatable],
        'type': item,
        'width': 2,
        'height': 2
    },
    'flamethrower-turret': {
        'mixins': [Rotatable],
        'type': item,
        'width': 2,
        'height': 3
    },


    'productivity-module': {
        'type': item
    },
    'productivity-module-2': {
        'type': item
    },
    'productivity-module-3': {
        'type': item
    },
    'effectivity-module': {
        'type': item
    },
    'effectivity-module-2': {
        'type': item
    },
    'effectivity-module-3': {
        'type': item
    },
    'speed-module': {
        'type': item
    },
    'speed-module-2': {
        'type': item
    },
    'speed-module-3': {
        'type': item
    },



    'water': {
        'type': fluid
    },
    'crude-oil': {
        'type': fluid
    },
    'petroleum-gas': {
        'type': fluid
    },
    'heavy-oil': {
        'type': fluid
    },
    'light-oil': {
        'type': fluid
    },
    'sulfuric-acid': {
        'type': fluid
    },
    'lubricant': {
        'type': fluid
    },
    'steam': {
        'type': fluid
    },

    'advanced-oil-processing': {
        'type': recipe
    },

    'raw-fish': {
        'type': item
    },
    'wood': {
        'type': item
    },
    'raw-wood': {
        'type': item
    },
    'iron-ore': {
        'type': item
    },
    'iron-plate': {
        'type': item
    },
    'copper-ore': {
        'type': item
    },
    'copper-plate': {
        'type': item
    },
    'steel-plate': {
        'type': item
    },
    'coal': {
        'type': item
    },
    'uranium-ore': {
        'type': item
    },
    'plastic-bar': {
        'type': item
    },
    'sulfur': {
        'type': item
    },

    'crude-oil-barrel': {
        'type': item
    },
    'heavy-oil-barrel': {
        'type': item
    },
    'light-oil-barrel': {
        'type': item
    },
    'lubricant-barrel': {
        'type': item
    },
    'petroleum-gas-barrel': {
        'type': item
    },
    'sulfuric-acid-barrel': {
        'type': item
    },
    'water-barrel': {
        'type': item
    },
    'empty-barrel': {
        'type': item
    },

    'processing-unit': {
        'type': item
    },

    'engine-unit': {
        'type': item
    },

    'electric-engine-unit': {
        'type': item
    },

    'battery': {
        'type': item
    },

    'explosives': {
        'type': item
    },
    'flying-robot-frame': {
        'type': item
    },
    'low-density-structure': {
        'type': item
    },
    'rocket-fuel': {
        'type': item
    },
    'rocket-control-unit': {
        'type': item
    },
    'satellite': {
        'type': item
    },
    'uranium-235': {
        'type': item
    },
    'uranium-238': {
        'type': item
    },

    'uranium-fuel-cell': {
        'type': item
    },
    'used-up-uranium-fuel-cell': {
        'type': item
    },
    'science-pack-1': {
        'type': item
    },
    'science-pack-2': {
        'type': item
    },
    'science-pack-3': {
        'type': item
    },
    'military-science-pack': {
        'type': item
    },
    'production-science-pack': {
        'type': item
    },
    'high-tech-science-pack': {
        'type': item
    },
    'space-science-pack': {
        'type': item
    },

    'iron-stick': {
        'type': item
    },
    'iron-gear-wheel': {
        'type': item
    },


    'signal-anything': {
        'type': virtual,
        'combinator': True
    },
    'signal-each': {
        'type': virtual,
        'combinator': True
    },
    'signal-everything': {
        'type': virtual,
        'combinator': True
    },
    'signal-0': {
        'type': virtual
    },
    'signal-1': {
        'type': virtual
    },
    'signal-2': {
        'type': virtual
    },
    'signal-3': {
        'type': virtual
    },
    'signal-4': {
        'type': virtual
    },
    'signal-5': {
        'type': virtual
    },
    'signal-6': {
        'type': virtual
    },
    'signal-7': {
        'type': virtual
    },
    'signal-8': {
        'type': virtual
    },
    'signal-9': {
        'type': virtual
    },
    'signal-A': {
        'type': virtual
    },
    'signal-B': {
        'type': virtual
    },
    'signal-C': {
        'type': virtual
    },
    'signal-D': {
        'type': virtual
    },
    'signal-E': {
        'type': virtual
    },
    'signal-F': {
        'type': virtual
    },
    'signal-G': {
        'type': virtual
    },
    'signal-H': {
        'type': virtual
    },
    'signal-I': {
        'type': virtual
    },
    'signal-J': {
        'type': virtual
    },
    'signal-K': {
        'type': virtual
    },
    'signal-L': {
        'type': virtual
    },
    'signal-M': {
        'type': virtual
    },
    'signal-N': {
        'type': virtual
    },
    'signal-O': {
        'type': virtual
    },
    'signal-P': {
        'type': virtual
    },
    'signal-Q': {
        'type': virtual
    },
    'signal-R': {
        'type': virtual
    },
    'signal-S': {
        'type': virtual
    },
    'signal-T': {
        'type': virtual
    },
    'signal-U': {
        'type': virtual
    },
    'signal-V': {
        'type': virtual
    },
    'signal-W': {
        'type': virtual
    },
    'signal-X': {
        'type': virtual
    },
    'signal-Y': {
        'type': virtual
    },
    'signal-Z': {
        'type': virtual
    },

    'signal-blue': {
        'type': virtual
    },
    'signal-red': {
        'type': virtual
    },
    'signal-green': {
        'type': virtual
    },
    'signal-yellow': {
        'type': virtual
    },
    'signal-cyan': {
        'type': virtual
    },
    'signal-pink': {
        'type': virtual
    },
    'signal-white': {
        'type': virtual
    },
    'signal-grey': {
        'type': virtual
    },
    'signal-black': {
        'type': virtual
    }
}



from py_factorio_blueprints.entity_mixins import *

entity_prototypes = {
    'accumulator': {
        'mixins': [Accumulator]
    },
    'artillery-turret': {
        'mixins': [Rotatable]
    },
    'beacon': {
        'mixins': [Items]
    },
    'boiler': {
        'mixins': [Rotatable]
    },
    'burner-generator': {
        'mixins': [Rotatable]
    },
    'arithmetic-combinator': {
        'mixins': [Arithmetic, Rotatable],
    },
    'decider-combinator': {
        'mixins': [Decider, Rotatable],
    },
    'constant-combinator': {
        'mixins': [ConstantCombinator, Rotatable]
    },
    'container': {},
    'logistic-container': {
        'mixins': [Requester],
    },
    'infinity-container': {
        'mixins': [InfinityContainer],
    },
    'assembling-machine': {
        'mixins': [Rotatable, Recipe],
    },
    'rocket-silo': {
        'mixins': [Silo],
    },
    'furnace': {},
    'electric-energy-interface': {},
    'electric-pole': {},
    'gate': {
        'mixins': [Rotatable]
    },
    'generator': {
        'mixins': [Rotatable]
    },
    'heat-interface': {},
    'heat-pipe': {},
    'inserter': {
        'mixins': [CircuitCondition, FilterInserter, Rotatable],
    },
    'lab': {},
    'lamp': {
        'mixins': [Lamp]
    },
    'land-mine': {},
    'market': {},
    'mining-drill': {
        'mixins': [MiningDrill, Rotatable]
    },
    'offshore-pump': {
        'mixins': [CircuitCondition, Rotatable]
    },
    'pipe': {},
    'infinity-pipe': {},
    'pipe-to-ground': {
        'mixins': [Rotatable]
    },
    'player-port': {},
    'power-switch': {
        'mixins': [CircuitCondition]
    },
    'programmable-speaker': {
        'mixins': [Speaker],
    },
    'pump': {
        'mixins': [Rotatable, CircuitCondition]
    },
    'radar': {},
    'curved-rail': {
        'mixins': [Rotatable]
    },
    'straight-rail': {
        'mixins': [Rotatable]
    },
    'rail-chain-signal': {
        'mixins': [ChainSignal, Rotatable]
    },
    'rail-signal': {
        'mixins': [RailSignal, Rotatable]
    },
    'reactor': {},
    'roboport': {
        'mixins': [Roboport]
    },
    'simple-entity': {},
    'simple-entity-with-owner': {},
    'simple-entity-with-force': {},
    'solar-panel': {},
    'storage-tank': {
        'mixins': [Rotatable]
    },
    'train-stop': {
        'mixins': [Rotatable, Station, Color]
    },
    'loader-1x1': {
        'mixins': [Rotatable]
    },
    'loader': {
        'mixins': [Rotatable, Loader],
    },
    'splitter': {
        'mixins': [Rotatable, Splitter]
    },
    'transport-belt': {
        'mixins': [Rotatable, TransportBelt, CircuitCondition],
    },
    'underground-belt': {
        'mixins': [Rotatable, Underground],
    },
    'turret': {
        'mixins': [Rotatable],
    },
    'ammo-turret': {
        'mixins': [Rotatable],
    },
    'electric-turret': {
        'mixins': [Rotatable],
    },
    'fluid-turret': {
        'mixins': [Rotatable],
    },
    'vehicle': {},
    'car': {},
    'artillery-wagon': {
        'mixins': [Train],
    },
    'cargo-wagon': {
        'mixins': [Cargo],
    },
    'fluid-wagon': {
        'mixins': [Train],
    },
    'locomotive': {
        'mixins': [Train, Items],
    },
    'wall': {},
}



class InvalidExchangeString(Exception):
    pass


class UnknownPrototype(Exception):
    pass


class UnknownEntity(UnknownPrototype):
    pass


class UnknownRecipe(UnknownPrototype):
    pass


class UnknownSignal(UnknownPrototype):
    pass


class UnknownItem(UnknownPrototype):
    pass


class UnknownTile(UnknownPrototype):
    pass


class DuplicateEntity(Exception):
    pass


class EntityOverlap(Exception):
    pass



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='py_factorio_blueprints',
    packages=['py_factorio_blueprints'],
    version='0.2.5',
    license='MIT',
    description='A python package to help create, '
                'modify and export factorio blueprints',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Tijmen Zwaan',
    author_email='tijmen.zwaan@gmail.com',
    url='https://github.com/tzwaan/python-factorio-blueprints',
    download_url='https://github.com/tzwaan/python-factorio-blueprints/archive/v0.2.5.tar.gz',
    keywords=['factorio', 'blueprint'],
    install_requires=[],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        'Natural Language :: English'
    ],
    scripts=[],
    include_package_data=True,
    zip_safe=False)
